use std::collections::hash_map::Entry;
use std::collections::{HashMap, HashSet};
use std::path::Path;

use color_eyre::eyre::{eyre, ContextCompat, WrapErr};
use color_eyre::Result;
use resvg::usvg::{TextRendering, TreeParsing, TreeTextToPath};
use tiny_skia::{Pixmap, Transform};

use crate::model::config::{Config, IconFormat, IconPack};
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::model::rgb::RGB8Wrapper;
use crate::model::IconMap;

mod filter;

#[derive(Debug)]
pub struct LoadedIcon {
    pub pixmap: Pixmap,
    pub scale: f32,
    pub clockwise_quarter_rotations: u8,
    pub border: Option<RGB8Wrapper>,
}

pub fn get_used_icon_descriptors(config: &Config) -> HashSet<IconDescriptor> {
    let mut result: HashSet<IconDescriptor> = HashSet::new();

    fn insert_all_from_map<T>(result: &mut HashSet<IconDescriptor>, map: &IconMap<T>) {
        map.values().for_each(|v| {
            result.insert(v.clone());
        });
    }

    for page in config.key_pages_by_id.values() {
        for key in page.keys.values() {
            if let Some(d) = &key.icon {
                result.insert(d.clone());

                if let Some(c) = &key.mode.media__next {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.media__play_pause {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.media__previous {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.home_assistant__button {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.home_assistant__switch {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.spotify__repeat {
                    insert_all_from_map(&mut result, &c.icon);
                }

                if let Some(c) = &key.mode.spotify__shuffle {
                    insert_all_from_map(&mut result, &c.icon);
                }
            }
        }
    }

    for page in config.knob_pages_by_id.values() {
        for knob in page.knobs.values() {
            result.insert(knob.icon.clone());

            if let Some(c) = &knob.mode.audio_volume {
                insert_all_from_map(&mut result, &c.icon);
            }
        }
    }

    result
}

pub fn load_icons(
    config_directory: &Path,
    icon_packs_by_id: &HashMap<String, IconPack>,
    descriptors: HashSet<IconDescriptor>,
    dpi: f32,
) -> Result<HashMap<IconDescriptor, LoadedIcon>> {
    let mut highest_scale_by_source: HashMap<IconDescriptorSource, f32> = HashMap::new();

    for d in &descriptors {
        let mut scale = d.filter.scale;
        if let IconDescriptorSource::IconPack { pack_id, .. } = &d.source {
            let pack = &icon_packs_by_id[pack_id];
            if let Some(filter) = &pack.global_filter {
                scale *= filter.scale;
            }
        }

        if let Some(v) = highest_scale_by_source.get_mut(&d.source) {
            if *v < scale {
                *v = scale;
            }
        } else {
            highest_scale_by_source.insert(d.source.clone(), scale);
        }
    }

    let mut unfiltered_scaled_pixmap_by_source: HashMap<IconDescriptorSource, (Pixmap, f32)> = HashMap::new();
    let mut icons_by_descriptor: HashMap<IconDescriptor, LoadedIcon> = HashMap::new();
    let mut fonts_db = resvg::usvg::fontdb::Database::new();
    fonts_db.load_system_fonts();

    for descriptor in descriptors {
        if descriptor.source == IconDescriptorSource::None {
            continue;
        }

        let (original_image, original_image_scale) = match unfiltered_scaled_pixmap_by_source.entry(descriptor.source.clone()) {
            Entry::Occupied(o) => o.into_mut(),
            Entry::Vacant(v) => v.insert(read_image_and_get_scale(
                config_directory,
                icon_packs_by_id,
                dpi,
                &fonts_db,
                descriptor.source.clone(),
                highest_scale_by_source[&descriptor.source],
            )?),
        };

        let (pixmap, scale) = if let IconDescriptorSource::IconPack { pack_id, .. } = &descriptor.source {
            let pack = &icon_packs_by_id[pack_id];
            if let Some(global_filter) = &pack.global_filter {
                (
                    filter::apply(&filter::apply(original_image, global_filter)?, &descriptor.filter)?,
                    descriptor.filter.scale * global_filter.scale,
                )
            } else {
                (filter::apply(original_image, &descriptor.filter)?, descriptor.filter.scale)
            }
        } else {
            (filter::apply(original_image, &descriptor.filter)?, descriptor.filter.scale)
        };

        let scale = scale / *original_image_scale;
        let clockwise_quarter_rotations = descriptor.filter.clockwise_quarter_rotations;
        let border = descriptor.filter.border;
        icons_by_descriptor.insert(
            descriptor,
            LoadedIcon {
                pixmap,
                scale,
                clockwise_quarter_rotations,
                border,
            },
        );
    }

    Ok(icons_by_descriptor)
}

fn read_image_and_get_scale(
    config_directory: &Path,
    icon_packs_by_id: &HashMap<String, IconPack>,
    dpi: f32,
    fonts_db: &resvg::usvg::fontdb::Database,
    source: IconDescriptorSource,
    highest_scale: f32,
) -> Result<(Pixmap, f32)> {
    let path = match source {
        IconDescriptorSource::None => return Ok((Pixmap::new(1, 1).unwrap(), 1.0)),
        IconDescriptorSource::Path(path) => path,
        IconDescriptorSource::IconPack { pack_id, icon_id } => {
            let pack = icon_packs_by_id.get(&pack_id).wrap_err_with(|| format!("Unknown icon pack: @{}", pack_id))?;

            let extension = match pack.format {
                IconFormat::Png => "png",
                IconFormat::Svg => "svg",
            };

            config_directory.join(&pack.path).join(icon_id + "." + extension)
        }
    };

    Ok(match path.extension() {
        None => return Err(eyre!("Invalid icon path: {:?}", path)),
        Some(extension) => match extension.to_string_lossy().as_ref() {
            "png" => (
                Pixmap::load_png(&path).wrap_err_with(|| format!("Failed to open or decode the PNG file at {}", path.to_string_lossy()))?,
                1.0,
            ),
            "svg" => (
                read_image_from_svg(&path, dpi, fonts_db, highest_scale)
                    .wrap_err_with(|| format!("Failed to open or decode the SVG file at {}", path.to_string_lossy()))?,
                highest_scale,
            ),
            extension => return Err(eyre!("Invalid file extension, only *.png and *.svg are allowed: {}", extension)),
        },
    })
}

fn read_image_from_svg(path: &Path, dpi: f32, font_db: &resvg::usvg::fontdb::Database, scale: f32) -> Result<Pixmap> {
    let raw_data = std::fs::read(path)?;

    let tree = {
        let mut tree = resvg::usvg::Tree::from_data(
            &raw_data,
            &resvg::usvg::Options {
                dpi,
                font_family: "Inter".to_owned(),
                font_size: 11.0,
                text_rendering: TextRendering::OptimizeLegibility,
                ..resvg::usvg::Options::default()
            },
        )?;

        tree.convert_text(font_db);

        resvg::Tree::from_usvg(&tree)
    };

    let size = tree.size.to_int_size();
    let mut pixmap = Pixmap::new((size.width() as f32 * scale).ceil() as u32, (size.height() as f32 * scale).ceil() as u32).unwrap();

    tree.render(Transform::from_scale(scale, scale), &mut pixmap.as_mut());

    Ok(pixmap)
}
