use std::fmt::{Debug, Display, Formatter};
use std::hash::{Hash, Hasher};
use std::str::FromStr;

use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;

use crate::model::geometry::IntRectWrapper;
use crate::model::rgb::RGB8Wrapper;

#[derive(Debug, PartialEq, Clone, SerializeDisplay, DeserializeFromStr)]
pub struct ImageFilter {
    pub crop: Option<IntRectWrapper>,
    pub scale: f32,
    pub clockwise_quarter_rotations: u8,
    pub color: Option<RGB8Wrapper>,
    pub alpha: f32,
    pub grayscale: bool,
    pub invert: bool,
    pub border: Option<RGB8Wrapper>,
}

impl Eq for ImageFilter {
    fn assert_receiver_is_total_eq(&self) {}
}

impl Hash for ImageFilter {
    fn hash<H: Hasher>(&self, hasher: &mut H) {
        self.to_string().hash(hasher)
    }
}

const DEFAULT_IMAGE_FILTER: ImageFilter = ImageFilter {
    scale: 1.0,
    clockwise_quarter_rotations: 0,
    crop: None,
    color: None,
    alpha: 1.0,
    grayscale: false,
    invert: false,
    border: None,
};

impl Default for ImageFilter {
    fn default() -> Self {
        DEFAULT_IMAGE_FILTER.clone()
    }
}

#[derive(Debug, Error)]
pub enum ImageFilterFromStringError {
    #[error("Unknown filter: {name}")]
    UnknownFilter { name: String },

    #[error("Filter {name} can only be used once.")]
    FilterUsedMoreThanOnce { name: String },

    #[error("Only the following values are allowed for rotate: -90, 90, 180, 270")]
    RotationNotAllowed,

    #[error("Filter {filter_name} requires a value ({filter_name}=<value>).")]
    FilterValueMissing { filter_name: String },

    #[error("Filter {filter_name} does not require or allow a value.")]
    FilterValueIgnored { filter_name: String },

    #[error("The following value supplied to {filter_name} could not be parsed: {raw_value}")]
    FilterValueNotParsable { filter_name: String, raw_value: String },

    #[error("{value} is not in the range ({range}) required by {filter_name}.")]
    FilterValueNotInRange { filter_name: String, value: String, range: String },
}

impl FromStr for ImageFilter {
    type Err = ImageFilterFromStringError;

    fn from_str<'a>(s: &str) -> Result<Self, Self::Err> {
        fn parse_rect_filter_value(filter_name: &str, raw_value: String) -> Result<IntRectWrapper, ImageFilterFromStringError> {
            IntRectWrapper::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                filter_name: filter_name.to_string(),
                raw_value,
            })
        }

        fn parse_f32_filter_value(
            filter_name: &str,
            raw_value: String,
            range_string: &str,
            check_range: Box<dyn Fn(&f32) -> bool>,
        ) -> Result<f32, ImageFilterFromStringError> {
            let value = f32::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                filter_name: filter_name.to_string(),
                raw_value,
            })?;

            if !check_range(&value) {
                return Err(ImageFilterFromStringError::FilterValueNotInRange {
                    filter_name: filter_name.to_string(),
                    range: range_string.to_string(),
                    value: value.to_string(),
                });
            }

            Ok(value)
        }

        let filters: Vec<&str> = s.split('|').map(|f| f.trim()).collect();

        let mut result = ImageFilter::default();
        let mut previous_filter_names: Vec<String> = Vec::new();

        for filter in filters {
            let split_filter = filter.split_once('=');
            let (filter_name, optional_raw_value) = if let Some((filter_name, raw_value)) = split_filter {
                (filter_name.to_owned(), Some(raw_value.to_owned()))
            } else {
                (filter.to_owned(), None)
            };

            if previous_filter_names.contains(&filter_name) {
                return Err(ImageFilterFromStringError::FilterUsedMoreThanOnce { name: filter_name });
            } else {
                previous_filter_names.push(filter_name.clone());
            }

            let value_is_present = optional_raw_value.is_some();
            let use_bool_value = || -> Result<bool, ImageFilterFromStringError> {
                if value_is_present {
                    Err(ImageFilterFromStringError::FilterValueIgnored {
                        filter_name: filter_name.clone(),
                    })
                } else {
                    Ok(true)
                }
            };

            let use_raw_value = || -> Result<String, ImageFilterFromStringError> {
                optional_raw_value.ok_or_else(|| ImageFilterFromStringError::FilterValueMissing {
                    filter_name: filter_name.clone(),
                })
            };

            match filter_name.as_str() {
                "crop" => result.crop = Some(parse_rect_filter_value(&filter_name, use_raw_value()?)?),
                "scale" => result.scale = parse_f32_filter_value(&filter_name, use_raw_value()?, "0..=100", Box::new(|v| (0.0..=100.0).contains(v)))?,
                "rotate" => {
                    let raw_value = use_raw_value()?;
                    let value = i32::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                        filter_name: filter_name.to_string(),
                        raw_value,
                    })?;

                    result.clockwise_quarter_rotations = match value {
                        0 => 0,
                        90 => 1,
                        180 => 2,
                        270 | -90 => 3,
                        _ => return Err(ImageFilterFromStringError::RotationNotAllowed),
                    };
                }
                "color" => {
                    let raw_value = use_raw_value()?;

                    result.color = Some(
                        RGB8Wrapper::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                            filter_name: filter_name.to_string(),
                            raw_value,
                        })?,
                    )
                }
                "alpha" => result.alpha = parse_f32_filter_value(&filter_name, use_raw_value()?, "0..<1", Box::new(|v| (0.0..1.0).contains(v)))?,
                "grayscale" => result.grayscale = use_bool_value()?,
                "invert" => result.invert = use_bool_value()?,
                "border" => {
                    let raw_value = use_raw_value()?;

                    result.border = Some(
                        RGB8Wrapper::from_str(&raw_value).map_err(|_| ImageFilterFromStringError::FilterValueNotParsable {
                            filter_name: filter_name.to_string(),
                            raw_value,
                        })?,
                    )
                }
                _ => return Err(ImageFilterFromStringError::UnknownFilter { name: filter_name }),
            };
        }

        Ok(result)
    }
}

impl Display for ImageFilter {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let mut is_first = true;

        if let Some(rect) = &self.crop {
            // if !is_first {
            //     f.write_str("|")?
            // }
            f.write_fmt(format_args!("crop={}", rect))?;
            is_first = false;
        }

        if self.scale != DEFAULT_IMAGE_FILTER.scale {
            if !is_first {
                f.write_str("|")?
            }
            f.write_fmt(format_args!("scale={}", self.scale))?;
            is_first = false;
        }

        if self.clockwise_quarter_rotations != DEFAULT_IMAGE_FILTER.clockwise_quarter_rotations {
            if !is_first {
                f.write_str("|")?
            }

            f.write_fmt(format_args!("rotate={}", self.clockwise_quarter_rotations as u16 * 90))?;
            is_first = false;
        }

        if let Some(color) = self.color {
            if !is_first {
                f.write_str("|")?
            }

            f.write_fmt(format_args!("color={}", color))?;
            is_first = false;
        }

        if self.alpha != DEFAULT_IMAGE_FILTER.alpha {
            if !is_first {
                f.write_str("|")?
            }
            f.write_fmt(format_args!("alpha={}", self.alpha))?;
            is_first = false;
        }

        if self.grayscale {
            if !is_first {
                f.write_str("|")?
            }
            f.write_str("grayscale")?;
            is_first = false;
        }

        if self.invert {
            if !is_first {
                f.write_str("|")?
            }
            f.write_str("invert")?;
            is_first = false;
        }

        if let Some(color) = self.color {
            if !is_first {
                f.write_str("|")?
            }

            f.write_fmt(format_args!("color={}", color))?;
            // is_first = false;
        }

        Ok(())
    }
}
